local silo = require("silo")

silo("server", function(ctx)
  local items = ctx:require("silo.items")
  local protocol = ctx:require("silo.protocol")
  local messaging = ctx:require("silo.messaging")
  local inventories = ctx:require("silo.inventories")

  local function broadcastItemUpdate()
    messaging:send({
      type = "update_items",
      inventories = items.inventories,
      itemCache = items.itemCache
    })
  end

  protocol:listen("storage.deposit_items", function(sender, message)
    for i, chest in pairs(message.chests) do
      local remote = peripheral.wrap(chest)

      while true do
        local count = 0

        local list = remote.list()
        for slot, _ in pairs(list) do
          local item = list[slot]
          if item == nil then break end

          count = count + 1
          item.slot = slot
          local entry = items:getItem(items.hashItem(item), remote, slot)
          items:insert(remote, entry, item)
          list = remote.list()
        end

        if count <= 0 then break end
      end
    end

    return {
      type = "storage.deposit_item_response",
      success = true
    }
  end)

  protocol:listen("storage.deposit_stack", function(sender, message)
    local from = peripheral.wrap(message.from)
    local fromSlot = message.fromSlot

    local item = from.list()[fromSlot]
    local fail = { type = "storage.deposit_stack_response", success = false }
    if item == nil then return fail end

    item.slot = fromSlot
    local entry = items:getItem(items.hashItem(item), from, fromSlot)
    if entry == nil then return fail end

    items:insert(from, entry, item)

    return {
      type = "storage.deposit_stack_response",
      success = true
    }
  end)

  protocol:listen("storage.withdraw_items", function(sender, message)
    local p = peripheral.wrap(message.to)
    for hash, count in pairs(message.items) do
      local entry = items:getItem(hash)
      items:extract(p, entry, count)
    end

    return {
      type = "storage.withdraw_item_response",
      success = true
    }
  end)

  protocol:listen("storage.withdraw_stack", function(sender, message)
    local to = peripheral.wrap(message.to)
    local toSlot = message.toSlot
    local hash = message.hash
    local count = message.count

    local entry = items:getItem(hash)
    if entry == nil then
      return {
        type = "storage.withdraw_stack_response",
        success = false
      }
    end

    items:extract(to, entry, count, toSlot)

    return {
      type = "storage.withdraw_stack_response",
      success = true
    }
  end)

  ctx.bus:register("peer.found", function(sender, message)
    broadcastItemUpdate()
  end)

  ctx.bus:register("items.change", function(change)
    broadcastItemUpdate()
  end)

  print("Starting...")
  inventories:scan(function()
    local totalSlots = 0
    local filledSlots = 0

    for name, inventory in pairs(items.inventories) do
      local slots = inventory.slots
      totalSlots = totalSlots + #slots

      for i, slot in pairs(slots) do
        if slot.hash ~= nil then
          filledSlots = filledSlots + 1
        end
      end
    end

    local pct = string.format("%.2f", 100 * (filledSlots / totalSlots))

    print(
      "Got " .. filledSlots .. " filled slots out of " ..
      totalSlots .. " (" .. pct .. "%)"
    )
  end)

  ctx:run()
end)
